// Direct.cpp
//
// Implementation for Direct3D support classes:
//
// CMatrix, CExecute

#include "stdafx.h"
#include "D3DEval.h"
#include "Direct.h"
#include <math.h>

#define D2R  0.01745329251994

////////////////////////////////////////////////////////////////////////
// CMatrix

CMatrix::CMatrix(IDirect3DDevice* pID3D)
{
    _11=D3DVAL(1.0); _12=D3DVAL(0.0);  _13=D3DVAL(0.0); _14=D3DVAL(0.0);
    _21=D3DVAL(0.0); _22=D3DVAL(1.0);  _23=D3DVAL(0.0); _24=D3DVAL(0.0);
    _31=D3DVAL(0.0); _32=D3DVAL(0.0);  _33=D3DVAL(1.0); _34=D3DVAL(0.0);
    _41=D3DVAL(0.0); _42=D3DVAL(0.0);  _43=D3DVAL(0.0); _44=D3DVAL(1.0);
	
	m_pID3D = NULL;
	m_hMatrix = NULL;
	if (pID3D) {
		// Create the matrix handle
		m_pID3D = pID3D;
		m_pID3D->AddRef();
		HRESULT hr = m_pID3D->CreateMatrix(&m_hMatrix);
		ASSERT(SUCCEEDED(hr));
		ASSERT(m_hMatrix);

		// Now set the initial matrix value
		Update();
	}
}

BOOL CMatrix::Create(IDirect3DDevice* pID3D)
{
	if (m_pID3D) return FALSE; // already created
	ASSERT(pID3D);
	m_pID3D = pID3D;
	m_pID3D->AddRef();
	HRESULT hr = m_pID3D->CreateMatrix(&m_hMatrix);
	ASSERT(SUCCEEDED(hr));
	ASSERT(m_hMatrix);

	// Now set the initial matrix value
	Update();
	return TRUE;
}

// virtual
CMatrix::~CMatrix()
{
#if 0 // BUGBUG Cant' usually delete these as the device
	// has already killed them by now
	if (m_hMatrix) {
		ASSERT(m_pID3D);
		m_pID3D->DeleteMatrix(m_hMatrix);
	}
	if (m_pID3D) {
		m_pID3D->Release();
	}
#endif
}

CMatrix::operator D3DMATRIXHANDLE() const
{
	ASSERT(m_hMatrix);
	return m_hMatrix;
}

// Update the physical matrix from the data values
void CMatrix::Update()
{
	if(m_hMatrix && m_pID3D) {
		HRESULT hr = m_pID3D->SetMatrix(m_hMatrix, this);
		ASSERT(SUCCEEDED(hr));
	}
}

////////////////////////////////////////////////////////////////////////
// CExecute

#define DEF_EXECUTE_SIZE 1024;

CExecute::CExecute(IDirect3DDevice* pID3D, IDirect3DViewport* pIViewport)
{
	// Save the device interface
	ASSERT(pID3D);
	m_pID3D = pID3D;
	m_pID3D->AddRef();

	// save the viewport interface
	ASSERT(pIViewport);
	m_pIViewport = pIViewport;
	m_pIViewport->AddRef();

	// Create an execute buffer of the default size
	memset(&m_Desc, 0, sizeof(m_Desc));
	m_Desc.dwSize = sizeof(m_Desc); // structure size
	m_Desc.dwFlags = D3DDEB_BUFSIZE; // dwBufferSize is valid
	m_Desc.dwBufferSize = DEF_EXECUTE_SIZE;
	HRESULT hr = m_pID3D->CreateExecuteBuffer(&m_Desc,
											  &m_pIEB,
											  NULL);
	ASSERT(SUCCEEDED(hr));
	ASSERT(m_pIEB);
	m_dwOffset = 0;

	// make sure we have at least an exit opcode in the buffer
	AddByte(D3DOP_EXIT);

	// set the offset back to the start
	m_dwOffset = 0;
	m_dwInsOffset = 0;
}

// virtual 
CExecute::~CExecute()
{
	if (m_pIViewport) {
		m_pIViewport->Release();
	}
	if (m_pIEB) {
		m_pIEB->Release();
	}
	if (m_pID3D) {
		m_pID3D->Release();
	}
}

// Lock the instruction buffer
void* CExecute::LockBuffer()
{
	if (m_Desc.lpData == NULL) {

		// Lock the buffer
		ASSERT(m_pIEB);
		HRESULT hr = m_pIEB->Lock(&m_Desc);
		ASSERT(SUCCEEDED(hr));
		ASSERT(m_Desc.lpData);
	}
	ASSERT(m_Desc.lpData);

	return m_Desc.lpData;
}

// Add a byte to the instruction buffer
BOOL CExecute::AddByte(BYTE b)
{
	// See if pointer to buffer exists
	if (m_Desc.lpData == NULL) {
		LockBuffer();
	}

	BYTE* pBuf = m_dwOffset + (BYTE*)m_Desc.lpData;
	*pBuf++ = b;
	m_dwOffset++;

	// Check that space remains in buffer
	ASSERT(m_dwOffset < m_Desc.dwBufferSize);

	// Append an exit code
	*pBuf = D3DOP_EXIT;

	return TRUE;
}

// Add a word to the instruction buffer
BOOL CExecute::AddStruct(void* pStruct, int iSize)
{
	ASSERT(pStruct);
	ASSERT(iSize > 0);
	BYTE* pSrc = (BYTE*) pStruct;
	while (iSize--) {
		if (!AddByte(*pSrc++)) return FALSE;
	}
	return TRUE;
}

// Add an instruction to the buffer
BOOL CExecute::AddIns(D3DOPCODE op, BYTE size, WORD count)
{
	D3DINSTRUCTION ins;
	ins.bOpcode = op;
	ins.bSize = size;
	ins.wCount = count;
	return AddStruct(&ins, sizeof(ins));
}

// Add a transform state matrix
BOOL CExecute::AddState(D3DTRANSFORMSTATETYPE t, D3DMATRIXHANDLE h)
{
	// Add the instruction header
	if (!AddIns(D3DOP_STATETRANSFORM, sizeof(D3DSTATE), 1)) {
		return FALSE;
	}
	
	// Add the data
	D3DSTATE s;
	s.dtstTransformStateType = t;
	s.dwArg[0] = h;
	return AddStruct(&s, sizeof(s));
}

// Add an ambient lighting command
BOOL CExecute::AddAmbLight(D3DCOLOR clr)
{
	// Add the instruction header
	if (!AddIns(D3DOP_STATELIGHT, sizeof(D3DSTATE), 1)) {
		return FALSE;
	}
	
	// Add the data
	D3DSTATE s;
	s.dlstLightStateType = D3DLIGHTSTATE_AMBIENT;
	s.dwArg[0] = clr;
	return AddStruct(&s, sizeof(s));
}

// Execute the current buffer
BOOL CExecute::Execute()
{
	// Make sure the buffer is unlocked
	ASSERT(m_pIEB);
	HRESULT hr;
	if (m_Desc.lpData != NULL) {

		// Unlock the buffer
		hr = m_pIEB->Unlock();
		ASSERT(SUCCEEDED(hr));
		m_Desc.lpData = NULL;
	}

	// Set up a buffer description
	// Note: we always assume that the vertices will be at the
	// start if there are any
	D3DEXECUTEDATA ed;
	memset(&ed, 0, sizeof(ed));
	ed.dwSize = sizeof(ed);
	ed.dwInstructionOffset = m_dwInsOffset; // offset to first instruction
	ed.dwInstructionLength = m_dwOffset - m_dwInsOffset + 1; // no. of instruction bytes
	ed.dwVertexCount = m_dwInsOffset / sizeof(D3DVERTEX);
	hr = m_pIEB->SetExecuteData(&ed);
	ASSERT(SUCCEEDED(hr));
	
	// Execute the buffer
	ASSERT(m_pID3D);
	hr = m_pID3D->BeginScene();
	ASSERT(SUCCEEDED(hr));
	hr = m_pID3D->Execute(m_pIEB, m_pIViewport, D3DEXECUTE_CLIPPED);
	BOOL b = SUCCEEDED(hr);
	hr = m_pID3D->EndScene();

	return b;
}

// Add a set of vertices to the buffer
// It is assumed that this will be the first command
// in a sequence
BOOL CExecute::AddVertices(D3DVERTEX* pVert, int nVert)
{
	ASSERT(m_dwOffset == 0); // we assume vertices will be first
	if (!AddStruct(pVert, nVert * sizeof(D3DVERTEX))) {
		return FALSE;
	}

	// Set the instruction offset value to just after the vertex list
	m_dwInsOffset = m_dwOffset;
	return TRUE;
}

// Add a set of transformed and lit vertices to the buffer
// It is assumed that this will be the first command
// in a sequence
BOOL CExecute::AddVertices(D3DTLVERTEX* pVert, int nVert)
{
	ASSERT(m_dwOffset == 0); // we assume vertices will be first
	if (!AddStruct(pVert, nVert * sizeof(D3DTLVERTEX))) {
		return FALSE;
	}

	// Set the instruction offset value to just after the vertex list
	m_dwInsOffset = m_dwOffset;
	return TRUE;
}

// Add a command to process the vertices in the buffer
BOOL CExecute::AddProcess(int nVert, int iStart, DWORD dwFlags)
{
	// Add the instruction header
	if (!AddIns(D3DOP_PROCESSVERTICES, sizeof(D3DPROCESSVERTICES), 1)) {
		return FALSE;
	}
	
	// Add the data
	D3DPROCESSVERTICES d;
	memset(&d, 0, sizeof(d));
	d.dwFlags = dwFlags;
	d.wStart = (WORD) iStart; // index of first vertex to process
	d.wDest =  d.wStart;
	d.dwCount = nVert;
	return AddStruct(&d, sizeof(d));
}

// Add a list of lines
BOOL CExecute::AddLines(D3DLINE* pLines, int nLines)
{
	// Add the instruction header
	if (!AddIns(D3DOP_LINE, sizeof(D3DLINE), nLines)) {
		return FALSE;
	}
	
	// Add the data
	return AddStruct(pLines, nLines * sizeof(D3DLINE));
}

// Add a single line
BOOL CExecute::AddLine(int v1, int v2)
{
	D3DLINE l;
	l.v1 = v1;
	l.v2 = v2;
	return AddLines(&l, 1);
}

BOOL CExecute::AddPoint(int v)
{
	D3DPOINT p;
	p.wCount = 1;
	p.wFirst = v;
	return AddPoints(&p);
}

BOOL CExecute::AddPoints(D3DPOINT* pPointInfo)
{
	// Add the instruction header
	if (!AddIns(D3DOP_POINT, sizeof(D3DPOINT), 1)) {
		return FALSE;
	}
	
	// Add the data
	return AddStruct(pPointInfo, sizeof(D3DPOINT));
}

BOOL CExecute::AddTriangle(int v1, int v2, int v3)
{
	// Add the instruction header
	if (!AddIns(D3DOP_TRIANGLE, sizeof(D3DTRIANGLE), 1)) {
		return FALSE;
	}
	
	// Add the data
	D3DTRIANGLE t;
	t.v1 = v1;
	t.v2 = v2;
	t.v3 = v3;
	t.wFlags = 0;
	return AddStruct(&t, sizeof(t));
}

BOOL CExecute::AddMaterial(D3DMATERIALHANDLE hMat)
{
	ASSERT(hMat);

	// Add the instruction header
	if (!AddIns(D3DOP_STATELIGHT, sizeof(D3DSTATE), 1)) {
		return FALSE;
	}

	// Add the data
	D3DSTATE s;
	s.dlstLightStateType = D3DLIGHTSTATE_MATERIAL;
	s.dwArg[0] = hMat;
	return AddStruct(&s, sizeof(s));
}

BOOL CExecute::GetExecuteData(D3DEXECUTEDATA* ped)
{
	ASSERT(ped);
	ASSERT(m_pIEB);
	HRESULT hr = m_pIEB->GetExecuteData(ped);
	return SUCCEEDED(hr);
}

BOOL CExecute::AddRenderState(D3DRENDERSTATETYPE state, DWORD value)
{
	// Add the instruction header
	if (!AddIns(D3DOP_STATERENDER, sizeof(D3DSTATE), 1)) {
		return FALSE;
	}

	// Add the data
	D3DSTATE s;
	s.drstRenderStateType = state;
	s.dwArg[0] = value;
	return AddStruct(&s, sizeof(s));
}

BOOL CExecute::AddMatMul(D3DMATRIXHANDLE hDest,
						   D3DMATRIXHANDLE hSrc1,
						   D3DMATRIXHANDLE hSrc2)
{
	ASSERT(hDest);
	ASSERT(hSrc1);
	ASSERT(hSrc2);

	// Add the instruction header
	if (!AddIns(D3DOP_MATRIXMULTIPLY, sizeof(D3DMATRIXMULTIPLY), 1)) {
		return FALSE;
	}

	// Add the data
	D3DMATRIXMULTIPLY mm;
	mm.hDestMatrix = hDest;
	mm.hSrcMatrix1 = hSrc1;
	mm.hSrcMatrix2 = hSrc2;
	return AddStruct(&mm, sizeof(mm));
}

////////////////////////////////////////////////////////////////////////
// CMaterial

CMaterial::CMaterial()
{
	// Initialize the data structure
	memset(&m_Mat, 0, sizeof(m_Mat));
	m_Mat.dwSize = sizeof(D3DMATERIAL);

	m_hMat = NULL;
	m_pIMat = NULL;
}

CMaterial::Create(IDirect3D* pIEngine, IDirect3DDevice* pIDevice)
{
	if (m_pIMat) return FALSE; // already created

	// Set the initial material state
	m_Mat.ambient.r = D3DVAL(1);
	m_Mat.ambient.g = D3DVAL(1);
	m_Mat.ambient.b = D3DVAL(1);
	m_Mat.ambient.a = D3DVAL(1);
	m_Mat.diffuse = m_Mat.ambient;
	m_Mat.specular = m_Mat.ambient;
 	m_Mat.dwRampSize = 16; // default shades

	// Create the material interface
	HRESULT hr = pIEngine->CreateMaterial(&m_pIMat, NULL);
	ASSERT(SUCCEEDED(hr));
	ASSERT(m_pIMat);

	// Set the current state of the material
	hr = m_pIMat->SetMaterial(&m_Mat);
	ASSERT(SUCCEEDED(hr));

	// Get a material handle for it from the device
	hr = m_pIMat->GetHandle(pIDevice, &m_hMat);
	ASSERT(SUCCEEDED(hr));
	ASSERT(m_hMat);
	return TRUE;
}

// virtual
CMaterial::~CMaterial()
{
	if (m_pIMat) {
		// BUGBUG: can't generally do this because device has killed it
		//m_pIMat->Release();
	}
}

CMaterial::operator D3DMATERIALHANDLE() const
{
	ASSERT(m_hMat);
	return m_hMat;
}

CMaterial::operator IDirect3DMaterial*() const
{
	ASSERT(m_pIMat);
	return m_pIMat;
}

BOOL CMaterial::SetColor(double r, double g, double b)
{
	m_Mat.ambient.r = D3DVAL(r);
	m_Mat.ambient.g = D3DVAL(g);
	m_Mat.ambient.b = D3DVAL(b);
	m_Mat.ambient.a = D3DVAL(1);
	m_Mat.diffuse.r = D3DVAL(r);
	m_Mat.diffuse.g = D3DVAL(g);
	m_Mat.diffuse.b = D3DVAL(b);
	m_Mat.diffuse.a = D3DVAL(1);
	m_Mat.specular.r = D3DVAL(1);
	m_Mat.specular.g = D3DVAL(1);
	m_Mat.specular.b = D3DVAL(1);
	m_Mat.specular.a = D3DVAL(1);
	m_Mat.power = D3DVAL(20.0);

	// Set the current state of the material
	ASSERT(m_pIMat);
	HRESULT hr = m_pIMat->SetMaterial(&m_Mat);
	return SUCCEEDED(hr);
}

BOOL CMaterial::SetShades(int n)
{
	m_Mat.dwRampSize = n;
	ASSERT(m_pIMat);
	HRESULT hr = m_pIMat->SetMaterial(&m_Mat);
	return SUCCEEDED(hr);
}

//////////////////////////////////////////////////////////////////////////
// CLight

CLight::CLight(IDirect3D* pIEngine)
{
	memset(&m_Light, 0, sizeof(m_Light));
	m_Light.dwSize = sizeof(m_Light);
	m_Light.dltType = D3DLIGHT_DIRECTIONAL;

	m_pILight = NULL;
	if (pIEngine) {
		HRESULT hr = pIEngine->CreateLight(&m_pILight, NULL);
		ASSERT(SUCCEEDED(hr));
		ASSERT(m_pILight);
	}
}

// virtual
CLight::~CLight()
{
	// BUGBUG the device killed it first
	if (m_pILight) {
		//m_pILight->Release();
	}
}

BOOL CLight::Create(IDirect3D* pIEngine)
{
	if (m_pILight) return FALSE; // already created
	ASSERT(pIEngine);
	HRESULT hr = pIEngine->CreateLight(&m_pILight, NULL);
	return SUCCEEDED(hr);
}

CLight::operator IDirect3DLight*() const
{
	ASSERT(m_pILight);
	return m_pILight;
}

void CLight::SetType(D3DLIGHTTYPE type)
{
	m_Light.dltType = type;
}

void CLight::SetColor(double r, double g, double b)
{
	D3DCOLORVALUE c;
	c.r = D3DVAL(r);
	c.g = D3DVAL(g);
	c.b = D3DVAL(b);
	c.a = D3DVAL(1);
	m_Light.dcvColor = c;
}

void CLight::SetDirection(double x, double y, double z)
{
	D3DVECTOR v;
	v.x = D3DVAL(x);
	v.y = D3DVAL(y);
	v.z = D3DVAL(z);
	m_Light.dvDirection = v;
}

void CLight::SetPosition(double x, double y, double z)
{
	D3DVECTOR v;
	v.x = D3DVAL(x);
	v.y = D3DVAL(y);
	v.z = D3DVAL(z);
	m_Light.dvPosition = v;
}

// Update the physical device from the local description
BOOL CLight::Update()
{
	ASSERT(m_pILight);
	HRESULT hr = m_pILight->SetLight(&m_Light);
	return SUCCEEDED(hr);
}